#include <stdio.h>
#include <stdarg.h>
#include "d7.h"
#include <math.h>

#define D7_PRINT_CALL 1
// combined impact
#define D7_PRINT_IMPACT 1
// poppers impact
#define D7_PRINT_IMPACT2 1


void d7_call( const char* format, ... ) {
	if ( D7_PRINT_CALL == 1)
	{
		va_list args;
		va_start( args, format );
		vfprintf( stderr, format, args );
		va_end( args );
	}
}




d7 dropdisc(d7 game, int col, int disc)
{
	d7_call("dropdisc(d7, %d, %d)\n", col, disc);

	int j = col;
	if ( 0 <= col && col <= 6 ) {
		if ( game.m[0][j] == 0 )
		{
			// replace first 0/empty spot from the bottom with value of disc
			for (int i = 6; i >= 0; --i)
			{
				if ( game.m[i][j] == 0 )	
				{
					game.m[i][j] = disc;
					break;
				}
			}
			game.status = D7_STATUS_CALCULATE;

		} else {
			printf("Not allowed! Colunm is full. \n");
			game.status = D7_STATUS_REQUIRE_DISC;
		}
	} else {
		printf("Not allowed! Select a column between 0 and 6. \n");
		game.status = D7_STATUS_REQUIRE_DISC;
	}
	return game;	
}


d7 calcturn(d7 game)
{
	d7_call("calcturn(d7)\n");

	// check columns
	d7_impact im1 = poppers_col(game);
	if ( D7_PRINT_IMPACT2 )
		printimpact(im1);

	// check rows
	d7_impact im2 = poppers_row(game);
	if ( D7_PRINT_IMPACT2 )
		printimpact(im2);

	// merge effects of rows and cols
	d7_impact im = merge_impact(im1, im2);
	if ( D7_PRINT_IMPACT )
		printimpact(im);
	

	// there is change so apply and test again
	if ( im.poppers != 0 )
	{

		// apply effects
		game = apply_impact(game, im, game.depth);
		
		// this makes sense as it shows the result ofter the other impact
		if ( D7_PRINT_IMPACT )
			printgrid(game);

		// apply gravity to discs
		game = gravity(game);
		if ( D7_PRINT_IMPACT )
			printgrid(game);
		
		// it may be that there is more to pop so test again
		game.depth++;
		game = calcturn(game);
	}


	return game;
}

d7_impact poppers_col(d7 game)
{
	d7_call("poppers_col(d7)\n");

	d7_impact im = {{{0}}};

	int chainlength;

	for (int j = 0; j < 7; ++j)
	{

		chainlength = 0;
		
		// find chainlength
		for (int i = 6; i >= 0; --i)
		{
			if ( game.m[i][j] == 0 ) 
				break;
			else 
				chainlength++;
		}	

		// pop discs and note down in impact matrix
		if ( chainlength > 0 ) 
		{
			// rather complicated for init to save cycles
			for (int i = 6; i >= 7-chainlength; --i)
			{
				if ( game.m[i][j] == chainlength ) {
					im.m[i][j] -= chainlength;
				
					// apply damages flags	
					if ( i < 6 )
						im.m[i+1][j] -= 10;
					
					if ( i > 0 ) 
						im.m[i-1][j] -= 100;
					
					if ( j < 6 )
						im.m[i][j+1] -= 1000;
					
					if ( j > 0 )
						im.m[i][j-1] -= 10000;
				
				} // if
			} // for		
		} // if
	} // for

	return im;
}



d7_impact poppers_row(d7 game)
{
	d7_call("poppers_row(d7)\n");

	d7_impact im = {{{0}}};
	int chainlength;
	int popnow;

	// iterate through the rows from bottom
	// for each row iterate from left to right

	// remember this is ROWS now ;)
	for (int i = 6; i >= 0; --i)
	{

		chainlength = 0;
		popnow = 0;
		
		for (int j = 0; j < 7; ++j)
		{

			if ( game.m[i][j] != 0 ) 
			{
				chainlength++;
				popnow = 0;
			}
			else
			{
				popnow = 1;
			} 


			if ( (popnow > 0 && chainlength > 0) ||  j==6 )
			{
				// if this is the last column the popcheck loop would not check
				// every field otherwise, changing j in the last run wont effect 
				// the outer for loop
				if ( j == 6 && popnow == 0 )
				{
					j = 7;
				}

				printf("--\n");
				for (int k = j-chainlength; k < j; ++k)
				{

					// if this pops note that in the impact matrix	
					if ( game.m[i][k] == chainlength )
					{

						im.m[i][k] -= chainlength;

						// apply damages flags	
						if ( i < 6 )
							im.m[i+1][k] -= 10;
						
						if ( i > 0 ) 
							im.m[i-1][k] -= 100;
						// take care for the next two, as k must be tested not j (in contrast to popcols)
						if ( k < 6 )
							im.m[i][k+1] -= 1000;
						
						if ( k > 0 )
							im.m[i][k-1] -= 10000;
						
					} // if
				} // for

				chainlength = 0;
				popnow = 0;

			} // if

		} // for
	} // for

	return im;
}

d7_impact merge_impact(d7_impact im1, d7_impact im2)
{
	d7_call("merge_impact(d7_impact, d7_impact)\n");
	
	int a,b,c,d, damages;

	for (int i = 0; i < 7; ++i)
	{
		for (int j = 0; j < 7; ++j)
		{
			// there are two strategies that might work
			// a) [implented here] goes easy on storage, but requires more condition tests
			// b) the other requires an the storage needed for d7_impace ( 7*7 sizeof int )

			if ( im1.m[i][j] % 10 < 0 
			  && im2.m[i][j] % 10 < 0 )
			{
				// it doesnt matter which to choose so we take value from im1 one
				// removes drops all damage info
				im1.m[i][j] = im1.m[i][j] % 10;
			}
			else 
			{
				im1.m[i][j] += im2.m[i][j];
			
				// as impacts are negative this is < test
				if ( im1.m[i][j] < 0 )
				{
					if ( im1.m[i][j] % 10 < 0 )
					{
						// this disc pops, so there is no need to have damages for this 
						// field
						im1.m[i][j] = im1.m[i][j] % 10;
						im1.poppers += 1;

					}
					else
					{
						// disc does not pop, but damages presented as flags so
						// convert them to actuall damage

						/* Example:
						 * lrtb
						 * abcd
						 * 12020
						 * ||||	
						 * |||+->    /10 = 1101 % 10 = 2
						 + ||+-->   /100 =  110 % 10 = 0
						 + |+--->  /1000 =   12 % 10 = 2
						 * +----> /10000 =             1
						 */	

						damages = 0;	
						a = im1.m[i][j]/10000;
						b = im1.m[i][j]/1000 % 10;
						c = im1.m[i][j]/100 % 10;
						d = im1.m[i][j]/10 % 10;

						/* f(0) = 0, f(x < 0) = -1,    0 > x > -4 
						 * there might be a better solution that does not require the
						 * floor() from math.h, as long as this is the only time
						 * a function from math.h is used this could be a point for
						 * optimisation.
						 */
						a =  floor(a / 4.0);	
						b =  floor(b / 4.0);	
						c =  floor(c / 4.0);	
						d =  floor(d / 4.0);	
					
						//printf("%d: %d, %d, %d, %d \n", im1.m[i][j], a, b, c, d);
						//printf("%d: %f, %f, %f, %f \n", im1.m[i][j], a, b, c, d);

						damages = a + b + c + d;


						/* fixing the damage matrix to represent only effectiv values
						 * seemed convienient but reducing damage to meaningful values
						 * needs to be done eventually at a later point anyways so it
						 * would be not very wise for speed.
						 */
						//if ( damages < -20 )
						//	damages = -20;
					
						im1.m[i][j] = 10*damages;
					}
				} // end if im1.m[i,j] < 0
			}


		}
	}
	return im1;
}


d7 apply_impact(d7 game, d7_impact im, int depth)
{
	d7_call("apply_impact(d7, d7_impact)\n");

	int substract = 0;

	for (int i = 0; i < 7; ++i)
	{
		for (int j = 0; j < 7; ++j)
		{
			substract = 0;

			if ( game.m[i][j] < 10 )
				substract = im.m[i][j] % 10;
			else
			{
				if ( im.m[i][j] <= -10 )
					substract -= 10;

				if ( im.m[i][j] <= -20 && game.m[i][j] >= 20 )
					substract -= 10;
			}
				

			game.m[i][j] += substract;
		}
	}

	game.score += im.poppers * (int)(7*pow(depth+1, 2.5));

	return game;
}

d7 gravity(d7 game)
{
	d7_call("gravity(d7)\n");

	int freefall;

	for (int j = 0; j < 7; ++j)
	{

		freefall = 0;

		for (int i = 6; i >= 0; --i)
		{
			
			if ( game.m[i][j] == 0 )
			{
				freefall++;
			}
			else if ( game.m[i][j] != 0 && freefall > 0 )
			{
				game.m[i+freefall][j] = game.m[i][j];
				game.m[i][j] = 0;
			}

		}
	}


	return game;
}


d7 levelup(d7 game, int* row) 
{
	d7_call("levelup(d7, row) \n");

	int toprow = 0;
	int bottomrow = 0;

	for (int j = 0; j < 7; ++j)
	{
		if ( game.m[0][j] != 0 ) {
			toprow++;
		}
	}

	for (int j = 0; j < 7; ++j)
	{
		if ( game.m[0][j] != 0 ) {
			bottomrow++;
		}
	}

	if ( toprow == 0)
	{
		
		for (int i = 1; i < 7; ++i)
		{
			for (int j = 0; j < 7; ++j)
			{
				game.m[i-1][j] = game.m[i][j];
				
				if (i == 6)
				{
					game.m[6][j] = row[j];
					if ( row[j] < 10 ) 
					{
						game.m[6][j] += 20;
					}
				}
			}
		}

		game.discs = game.nextleveldiscs;
		game.nextleveldiscs--;
		if (game.status != D7_STATUS_START_GAME )
		{
			game.score += 7000;
		}

	} else {
		game.status = D7_STATUS_GAME_OVER;
		printf("GAME OVER! \n");
		printf("GAME OVER! \n");
		printf("GAME OVER! \n");
	}

	return game;
}



//	nice format the d7 datatype and print it to console
void printgrid(d7 game) {
	for (int i = 0; i < 7; ++i)
	{
		for (int j = 0; j < 7; ++j)
		{
				printf("%d,\t", game.m[i][j]);
		}
		printf("\n");
	}
	printf("Score: %d", game.score);
	printf("   Level: %d", game.level);
	printf("   Turn: %d", game.turn);
	printf("   Status: %d", game.status);
	printf("   Discs: %d", game.discs);
	printf("   Depth: %d", game.depth);
	printf("\n\n");
}

//	nice format the d7 datatype and print it to console
void printimpact(d7_impact game) {
	for (int i = 0; i < 7; ++i)
	{
		printf("> ");
		for (int j = 0; j < 7; ++j)
		{
			printf("%d,\t", game.m[i][j]);
		}
		printf("\n");
	}

	printf("Poppers: %d", game.poppers);
	printf("\n\n");
}
